import abc
import functools
import operator
import os
import typing
import urllib.parse
from typing import Dict, Optional, Set

from ..logic.lineage import Lineage
from ..ontology import Ontology
from ..utils.pool import ThreadPool


class BaseParser(abc.ABC):
    def __init__(self, ont: Ontology):
        self.ont: Ontology = ont

    @classmethod
    @abc.abstractmethod
    def can_parse(cls, path: str, buffer: bytes) -> bool:
        """Return `True` if this parser type can parse the given handle."""
        return NotImplemented  # type: ignore

    @abc.abstractmethod
    def parse_from(
        self, handle: typing.BinaryIO, threads: Optional[int] = None
    ) -> None:
        return NotImplemented

    @classmethod
    def pool(cls, threads: int) -> ThreadPool:
        return ThreadPool(threads)

    @classmethod
    def process_import(
        cls,
        ref: str,
        import_depth: int = -1,
        basepath: str = "",
        timeout: int = 5,
    ) -> Ontology:
        s = urllib.parse.urlparse(ref).scheme
        if s in {"ftp", "http", "https"} or os.path.exists(ref):
            url = ref
        else:
            for ext in ["", ".obo", ".json", ".owl"]:
                if os.path.exists(os.path.join(basepath, f"{ref}{ext}")):
                    url = os.path.join(basepath, f"{ref}{ext}")
                    break
            else:
                id_ = f"{ref}.obo" if not os.path.splitext(ref)[1] else ref
                url = f"http://purl.obolibrary.org/obo/{id_}"
        return Ontology(url, max(import_depth - 1, -1), timeout)

    @classmethod
    def process_imports(
        cls,
        imports: Set[str],
        import_depth: int = -1,
        basepath: str = "",
        timeout: int = 5,
        threads: Optional[int] = None,
    ) -> Dict[str, Ontology]:
        # check we did not reach the maximum import depth
        if import_depth == 0:
            return dict()
        process = functools.partial(
            cls.process_import,
            import_depth=import_depth,
            basepath=basepath,
            timeout=timeout,
        )
        with cls.pool(threads) as pool:
            return dict(pool.map(lambda i: (i, process(i)), imports))

    _entities = {
        "Term": operator.attrgetter("terms", "_terms"),
        "Relationship": operator.attrgetter("relationships", "_relationships"),
    }

    def symmetrize_lineage(self):
        for getter in self._entities.values():
            entities, graphdata = getter(self.ont)
            for entity in entities():
                graphdata.lineage.setdefault(entity.id, Lineage())
            for subentity, lineage in graphdata.lineage.items():
                for superentity in lineage.sup:
                    graphdata.lineage[superentity].sub.add(subentity)

    def import_lineage(self):
        for getter in self._entities.values():
            entities, graphdata = getter(self.ont)
            for dep in self.ont.imports.values():
                dep_entities, dep_graphdata = getter(dep)
                for entity in dep_entities():
                    graphdata.lineage[entity.id] = Lineage()
                for id, lineage in dep_graphdata.lineage.items():
                    graphdata.lineage[id].sup.update(lineage.sup)
                    graphdata.lineage[id].sub.update(lineage.sub)



import datetime
import operator
import typing
from typing import Dict, Iterable, FrozenSet, Optional, Set, Tuple, Iterator

from .definition import Definition
from .entity import Entity, EntityData, EntitySet
from .logic import SubpropertiesIterator, SuperpropertiesIterator
from .logic.lineage import SubpropertiesHandler, SuperpropertiesHandler
from .pv import PropertyValue
from .synonym import SynonymData
from .utils.meta import typechecked
from .xref import Xref

if typing.TYPE_CHECKING:
    from .ontology import Ontology
    from .term import Term


__all__ = ["Relationship", "RelationshipData", "RelationshipSet"]


class RelationshipData(EntityData):
    """Internal data storage of `Relationship` information."""

    id: str
    anonymous: bool
    name: Optional[str]
    namespace: Optional[str]
    alternate_ids: Set[str]
    definition: Optional[Definition]
    comment: Optional[str]
    subsets: Set[str]
    synonyms: Set[SynonymData]
    xrefs: Set[Xref]
    annotations: Set[PropertyValue]
    domain: Optional[str]
    range: Optional[str]
    builtin: bool
    holds_over_chain: Set[Tuple[str, str]]
    antisymmetric: bool
    cyclic: bool
    reflexive: bool
    asymmetric: bool
    symmetric: bool
    transitive: bool
    functional: bool
    inverse_functional: bool
    intersection_of: Set[str]
    inverse_of: Optional[str]
    transitive_over: Set[str]
    equivalent_to_chain: Set[Tuple[str, str]]
    disjoint_over: Set[str]
    obsolete: bool
    created_by: Optional[str]
    creation_date: Optional[datetime.datetime]
    expand_assertion_to: Set[Definition]
    expand_expression_to: Set[Definition]
    metadata_tag: bool
    class_level: bool

    if typing.TYPE_CHECKING:
        __annotations__: Dict[str, str]

    __slots__ = tuple(__annotations__)  # noqa: E0602

    def __init__(
        self,
        id: str,
        anonymous: bool = False,
        name: Optional[str] = None,
        namespace: Optional[str] = None,
        alternate_ids: Optional[Set[str]] = None,
        definition: Optional[Definition] = None,
        comment: Optional[str] = None,
        subsets: Optional[Set[str]] = None,
        synonyms: Optional[Set[SynonymData]] = None,
        xrefs: Optional[Set[Xref]] = None,
        annotations: Optional[Set[PropertyValue]] = None,
        domain: Optional[str] = None,
        range: Optional[str] = None,
        builtin: bool = False,
        holds_over_chain: Optional[Set[Tuple[str, str]]] = None,
        antisymmetric: bool = False,
        cyclic: bool = False,
        reflexive: bool = False,
        asymmetric: bool = False,
        symmetric: bool = False,
        transitive: bool = False,
        functional: bool = False,
        inverse_functional: bool = False,
        intersection_of: Optional[Set[str]] = None,
        union_of: Optional[Set[str]] = None,
        equivalent_to: Optional[Set[str]] = None,
        disjoint_from: Optional[Set[str]] = None,
        inverse_of: Optional[str] = None,
        transitive_over: Optional[Set[str]] = None,
        equivalent_to_chain: Optional[Set[Tuple[str, str]]] = None,
        disjoint_over: Optional[Set[str]] = None,
        relationships: Optional[Dict[str, Set[str]]] = None,
        obsolete: bool = False,
        created_by: Optional[str] = None,
        creation_date: Optional[datetime.datetime] = None,
        replaced_by: Optional[Set[str]] = None,
        consider: Optional[Set[str]] = None,
        expand_assertion_to: Optional[Set[Definition]] = None,
        expand_expression_to: Optional[Set[Definition]] = None,
        metadata_tag: bool = False,
        class_level: bool = False,
    ):
        self.id = id
        self.anonymous = anonymous
        self.name = name
        self.namespace = namespace
        self.alternate_ids = alternate_ids or set()
        self.definition = definition
        self.comment = comment
        self.subsets = subsets or set()
        self.synonyms = synonyms or set()
        self.xrefs = xrefs or set()
        self.annotations = annotations or set()
        self.domain = domain
        self.range = range
        self.builtin = builtin
        self.holds_over_chain = holds_over_chain or set()
        self.antisymmetric = antisymmetric
        self.cyclic = cyclic
        self.reflexive = reflexive
        self.asymmetric = asymmetric
        self.symmetric = symmetric
        self.transitive = transitive
        self.functional = functional
        self.inverse_functional = inverse_functional
        self.intersection_of = intersection_of or set()
        self.union_of = union_of or set()
        self.equivalent_to = equivalent_to or set()
        self.disjoint_from = disjoint_from or set()
        self.inverse_of = inverse_of
        self.transitive_over = transitive_over or set()
        self.equivalent_to_chain = equivalent_to_chain or set()
        self.disjoint_over = disjoint_over or set()
        self.relationships = relationships or dict()
        self.obsolete = obsolete
        self.created_by = created_by
        self.creation_date = creation_date
        self.replaced_by = replaced_by or set()
        self.consider = consider or set()
        self.expand_assertion_to = expand_assertion_to or set()
        self.expand_expression_to = expand_expression_to or set()
        self.metadata_tag = metadata_tag
        self.class_level = class_level


class RelationshipSet(EntitySet["Relationship"]):
    """A specialized mutable set to store `Relationship` instances."""

    # --- Magic methods ------------------------------------------------------

    def __iter__(self) -> Iterator["Relationship"]:
        return map(lambda t: self._ontology.get_relationship(t), iter(self._ids))  # type: ignore

    # --- Methods ---------------------------------------------------------

    def subproperties(
        self, distance: Optional[int] = None, with_self: bool = True
    ) -> SubpropertiesIterator:
        """Get an iterator over the subproperties of all relationships in the set."""
        return SubpropertiesIterator(*self, distance=distance, with_self=with_self)

    def superproperties(
        self, distance: Optional[int] = None, with_self: bool = True
    ) -> SuperpropertiesIterator:
        """Get an iterator over the superproperties of all relationships in the set.

        Example:
            >>> pato = pronto.Ontology("pato.obo")
            >>> proportionality_to = pato["PATO:0001470"]
            >>> quality_mapping = pronto.RelationshipSet(
            ...     r for r in pato.relationships()
            ...     if r.domain == proportionality_to
            ... )
            >>> sorted(quality_mapping.subproperties().to_set().ids)
            ['has_dividend_entity', 'has_dividend_quality', ...

        """
        return SuperpropertiesIterator(*self, distance=distance, with_self=with_self)


class Relationship(Entity["RelationshipData", "RelationshipSet"]):
    """A relationship, constitute the edges of the ontology graph.

    Also sometimes refered as typedefs, relationship types, properties or
    predicates. Formally equivalent to a property (either ``ObjectProperty``
    or ``AnnotationProperty``) in OWL2.
    """

    if typing.TYPE_CHECKING:

        def __init__(self, ontology: "Ontology", reldata: "RelationshipData"):
            super().__init__(ontology, reldata)

        def _data(self) -> "RelationshipData":
            return typing.cast("RelationshipData", super()._data())

    # --- Associated type variables ------------------------------------------

    _Set = RelationshipSet
    _data_getter = operator.attrgetter("_relationships")

    # --- Methods ------------------------------------------------------------

    def subproperties(
        self, distance: Optional[int] = None, with_self: bool = True
    ) -> "SubpropertiesHandler":
        """Get an handle over the subproperties of this `Relationship`.

        Arguments:
            distance (int, optional): The maximum distance between this
                relationship and the yielded subproperties (`0` for the
                relationship itself, `1` for its immediate children, etc.).
                Use `None` to explore the entire directed graph transitively.
            with_self (bool): Whether or not to include the current term in
                the terms being yielded. RDF semantics state that the
                ``rdfs:subClassOf`` property is reflexive (and therefore is
                ``rdfs:subPropertyOf`` reflexive too by transitivity), so this
                is enabled by default, but in most practical cases only the
                distinct subproperties are desired.

        """
        return SubpropertiesHandler(self, distance=distance, with_self=with_self)

    def superproperties(
        self, distance: Optional[int] = None, with_self: bool = True
    ) -> "SuperpropertiesHandler":
        """Get an handle over the superproperties of this `Relationship`.

        In order to follow the semantics of ``rdf:subPropertyOf``, which in
        turn respects the mathematical definition of subset inclusion, ``is_a``
        is defined as a transitive relationship, hence the inverse relationship
        is also transitive by closure property.

        Arguments:
            distance (int, optional): The maximum distance between this
                relationship and the yielded subperoperties (`0` for the
                relationship itself, `1` for its immediate parents, etc.).
                Use `None` to explore the entire directed graph transitively.
            with_self (bool): Whether or not to include the current term in
                the terms being yielded. RDF semantics state that the
                ``rdfs:subClassOf`` property is transitive (and therefore is
                ``rdfs:subPropertyOf`` transitive too), so this is enabled
                by default, but in most practical cases only the distinct
                subproperties are desired.

        """
        return SuperpropertiesHandler(self, distance=distance, with_self=with_self)

    # --- Attributes ---------------------------------------------------------

    @property
    def antisymmetric(self) -> bool:
        """`bool`: Whether this relationship is anti-symmetric."""
        return self._data().antisymmetric

    @antisymmetric.setter  # type: ignore
    @typechecked(property=True)
    def antisymmetric(self, value: bool) -> None:
        self._data().antisymmetric = value

    @property
    def asymmetric(self) -> bool:
        """`bool`: Whether this relationship is asymmetric."""
        return self._data().asymmetric

    @asymmetric.setter  # type: ignore
    @typechecked(property=True)
    def asymmetric(self, value: bool) -> None:
        self._data().asymmetric = value

    @property
    def class_level(self) -> bool:
        """`bool`: Whether this relationship is applied at class level.

        This tag affects how OBO ``relationship`` tags should be translated
        in OWL2: by default, all relationship tags are taken to mean an
        all-some relation over an instance level relation. With this flag
        set to `True`, the relationship will be translated to an `owl:hasValue`
        restriction.
        """
        return self._data().class_level

    @class_level.setter  # type: ignore
    @typechecked(property=True)
    def class_level(self, value: bool) -> None:
        self._data().class_level = value

    @property
    def cyclic(self) -> bool:
        """`bool`: Whether this relationship is cyclic."""
        return self._data().cyclic

    @cyclic.setter  # type: ignore
    @typechecked(property=True)
    def cyclic(self, value: bool) -> None:
        self._data().cyclic = value

    @property
    def disjoint_over(self) -> "RelationshipSet":
        """`frozenset`: The relationships this relationships is disjoint over."""
        s = RelationshipSet()
        s._ids = self._data().disjoint_over
        s._ontology = self._ontology()
        return s

    @property
    def domain(self) -> Optional["Term"]:
        """`Term` or `None`: The domain of the relationship, if any."""
        data, ontology = self._data(), self._ontology()
        if data.domain is not None:
            return ontology.get_term(data.domain)
        return None

    @domain.setter
    def domain(self, value: Optional["Term"]) -> None:
        rshipdata, ontology = self._data(), self._ontology()
        if value is not None:
            try:
                ontology.get_term(value.id)
            except KeyError:
                raise ValueError(f"{value} is not a term in {ontology}")
        rshipdata.domain = value.id if value is not None else None

    @property
    def equivalent_to_chain(self) -> FrozenSet[Tuple["Relationship", "Relationship"]]:
        return frozenset(
            {
                tuple(map(self._ontology().get_relationship, chain))
                for chain in self._data().equivalent_to_chain
            }
        )

    @equivalent_to_chain.setter
    def equivalent_to_chain(self, equivalent_to_chain: Iterable[Tuple["Relationship", "Relationship"]]):
        data = self._data()
        data.equivalent_to_chain = {
            (r1.id, r2.id)
            for r1, r2 in equivalent_to_chain
        }

    @property
    def expand_assertion_to(self) -> FrozenSet[Definition]:
        return frozenset(self._data().expand_assertion_to)

    @property
    def expand_expression_to(self) -> FrozenSet[Definition]:
        return frozenset(self._data().expand_expression_to)

    @property
    def functional(self) -> bool:
        """`bool`: Whether this relationship is functional."""
        return self._data().functional

    @functional.setter  # type: ignore
    @typechecked(property=True)
    def functional(self, value: bool) -> None:
        self._data().functional = value

    @property
    def inverse_functional(self) -> bool:
        """`bool`: Whether this relationship is inverse functional."""
        return self._data().inverse_functional

    @inverse_functional.setter  # type: ignore
    @typechecked(property=True)
    def inverse_functional(self, value: bool) -> None:
        self._data().inverse_functional = value

    @property
    def metadata_tag(self) -> bool:
        """`bool`: Whether or not this relationship is a metadata tag.

        This tag affects how OBO typedefs should be translated in OWL2: by
        default, all typedef tags are translated to an `owl:ObjectProperty`.
        With this flag set to `True`, the typedef will be translated to an
        `owl:AnnotationProperty`.
        """
        return self._data().metadata_tag

    @metadata_tag.setter  # type: ignore
    @typechecked(property=True)
    def metadata_tag(self, value: bool):
        self._data().metadata_tag = value

    @property
    def holds_over_chain(self) -> FrozenSet[Tuple["Relationship", "Relationship"]]:
        """`frozenset` of `Relationship` couples: The chains this relationship holds over."""
        ont: "Ontology" = self._ontology()
        data: "RelationshipData" = self._data()
        return frozenset(
            tuple(map(ont.get_term, chain))
            for chain in data.holds_over_chain
        )

    @holds_over_chain.setter
    def holds_over_chain(self, holds_over_chain: Iterable[Tuple["Relationship", "Relationship"]]) -> None:
        data: "RelationshipData" = self._data()
        data.holds_over_chain = {
            (r1.id, r2.id)
            for r1, r2 in holds_over_chain
        }

    @property
    def inverse_of(self) -> Optional["Relationship"]:
        """`Relationship` or `None`: The inverse of this relationship, if any."""
        ont, reldata = self._ontology(), self._data()
        if reldata.inverse_of is not None:
            return ont.get_relationship(reldata.inverse_of)
        return None

    @inverse_of.setter
    def inverse_of(self, value: Optional["Relationship"]):
        self._data().inverse_of = None if value is None else value.id

    @property
    def intersection_of(self) -> "RelationshipSet":
        """`RelationshipSet`: The relations this relationship is an intersection of."""
        s = RelationshipSet()
        s._ids = self._data().intersection_of
        s._ontology = self._ontology()
        return s

    @property
    def range(self) -> Optional["Term"]:
        """`Term` or `None`: The range of the relationship, if any."""
        range, ont = self._data().range, self._ontology()
        return ont.get_term(range) if range is not None else None

    @range.setter
    def range(self, value: Optional["Term"]):
        if value is not None:
            try:
                self._ontology().get_term(value.id)
            except KeyError:
                raise ValueError(f"{value} is not in {self._ontology()}")
        self._data().range = value.id if value is not None else None

    @property
    def reflexive(self) -> bool:
        """`bool`: Whether or not the relationship is reflexive."""
        return self._data().reflexive

    @reflexive.setter  # type: ignore
    @typechecked(property=True)
    def reflexive(self, value: bool):
        self._data().reflexive = value

    @property
    def symmetric(self) -> bool:
        """`bool`: Whether or not the relationship is symmetric."""
        return self._data().symmetric

    @symmetric.setter  # type: ignore
    @typechecked(property=True)
    def symmetric(self, value: bool):
        self._data().symmetric = value

    @property
    def transitive(self) -> bool:
        """`bool`: Whether or not the relationship is transitive."""
        return self._data().transitive

    @transitive.setter  # type: ignore
    @typechecked(property=True)
    def transitive(self, value: bool):
        self._data().transitive = value

    @property
    def transitive_over(self) -> "RelationshipSet":
        """`RelationshipSet`: The relations this relationship is transitive over."""
        s = RelationshipSet()
        s._ids = self._data().transitive_over
        s._ontology = self._ontology()
        return s


# TODO: remove in v3.0.0
_BUILTINS = {
    "is_a": RelationshipData(
        id="is_a",
        anonymous=False,
        name="is a",
        namespace=None,
        alternate_ids=None,
        definition=Definition(
            "A subclassing relationship between one term and another",
            xrefs=set(
                {
                    Xref(
                        "http://owlcollab.github.io/oboformat/doc/GO.format.obo-1_4.html"
                    )
                }
            ),
        ),
        comment=None,
        subsets=None,
        synonyms=None,
        xrefs=None,
        annotations=None,
        domain=None,
        range=None,
        builtin=True,
        holds_over_chain=None,
        antisymmetric=True,
        cyclic=True,
        reflexive=True,
        asymmetric=False,
        symmetric=False,
        transitive=True,
        functional=False,
        inverse_functional=False,
        intersection_of=None,
        union_of=None,
        equivalent_to=None,
        disjoint_from=None,
        inverse_of=None,
        transitive_over=None,
        equivalent_to_chain=None,
        disjoint_over=None,
        relationships=None,
        obsolete=False,
        created_by=None,
        creation_date=None,
        replaced_by=None,
        consider=None,
        expand_assertion_to=None,  # TODO
        expand_expression_to=None,  # TODO
        metadata_tag=False,
        class_level=True,
    )
}



import typing
from typing import Callable, Iterable, List

try:
    from multiprocessing.pool import ThreadPool as _ThreadPool
except ImportError:
    _ThreadPool = None  # type: ignore


_T = typing.TypeVar("_T")
_U = typing.TypeVar("_U")


class ThreadPool(object):

    def __init__(self, threads: int = 0):
        self.threads = threads
        self.pool = None if _ThreadPool is None else _ThreadPool(self.threads)

    def __enter__(self) -> "Pool":
        if self.pool is not None:
            self.pool.__enter__()
        return self

    def __exit__(self, exc_val, exc_ty, tb):
        if self.pool is not None:
            return self.pool.__exit__(exc_val, exc_ty, tb)
        return False

    def map(self, func: Callable[[_T], _U], items: Iterable[_T]) -> List[_U]:
        if self.pool is None:
            return list(map(func, items))
        else:
            return self.pool.map(func, items)




from typing import BinaryIO

import fastobo

from ._fastobo import FastoboSerializer
from .base import BaseSerializer


class OwlFunctionalSerializer(FastoboSerializer, BaseSerializer):

    format = "ofn"

    def dump(self, file: BinaryIO):
        doc = self._to_obodoc(self.ont)
        fastobo.dump_owl(doc, file, format="ofn")



"""Warnings raised by the library.
"""


class ProntoWarning(Warning):
    """The class for all warnings raised by `pronto`."""

    pass


class NotImplementedWarning(ProntoWarning, NotImplementedError):
    """Some part of the code is yet to be implemented."""

    pass


class UnstableWarning(ProntoWarning):
    """The behaviour of the executed code might change in the future."""

    pass


class SyntaxWarning(ProntoWarning, SyntaxError):
    """The parsed document contains incomplete or unsound constructs."""

    def __init__(self, *args, **kwargs):
        ProntoWarning.__init__(self, *args, **kwargs)
        SyntaxError.__init__(self, *args, **kwargs)



import os

import fastobo

from ..logic.lineage import Lineage
from ..utils.meta import typechecked
from ._fastobo import FastoboParser
from .base import BaseParser


class OboJSONParser(FastoboParser, BaseParser):
    @classmethod
    def can_parse(cls, path, buffer):
        return buffer.lstrip().startswith(b"{")

    def parse_from(self, handle, threads=None):
        # Load the OBO graph into a syntax tree using fastobo
        doc = fastobo.load_graph(handle).compact_ids()

        # Extract metadata from the OBO header
        with typechecked.disabled():
            self.ont.metadata = self.extract_metadata(doc.header)

        # Resolve imported dependencies
        self.ont.imports.update(
            self.process_imports(
                self.ont.metadata.imports,
                self.ont.import_depth,
                os.path.dirname(self.ont.path or str()),
                self.ont.timeout,
                threads=threads,
            )
        )

        # Merge lineage cache from imports
        self.import_lineage()

        # Extract frames from the current document.
        with typechecked.disabled():
            try:
                with self.pool(threads) as pool:
                    pool.map(self.extract_entity, doc)
            except SyntaxError as err:
                location = self.ont.path, err.lineno, err.offset, err.text
                raise SyntaxError(err.args[0], location) from None

            # OBOJSON can define classes implicitly using only `is_a` properties
            # mapping to unresolved identifiers: in this case, we create the
            # term ourselves
            for lineage in list(self.ont._terms.lineage.values()):
                for superclass in lineage.sup.difference(self.ont._terms.lineage):
                    self.ont.create_term(superclass)

        # Update lineage cache
        self.symmetrize_lineage()



import datetime
import functools
import warnings
from operator import attrgetter
from typing import Union

import fastobo

from ..definition import Definition
from ..metadata import Metadata, Subset
from ..ontology import Ontology
from ..pv import LiteralPropertyValue, PropertyValue, ResourcePropertyValue
from ..relationship import Relationship, RelationshipData
from ..synonym import Synonym, SynonymData, SynonymType
from ..term import Term, TermData
from ..utils.warnings import NotImplementedWarning, SyntaxWarning
from ..xref import Xref

# --- Type Hint --------------------------------------------------------------

DefClause = Union[fastobo.term.DefClause, fastobo.typedef.DefClause]


# --- Parser interface -------------------------------------------------------


class FastoboParser:

    ont: "Ontology"

    # --- Doc processing -----------------------------------------------------

    def extract_entity(self, frame: fastobo.abc.AbstractEntityFrame) -> None:
        if isinstance(frame, fastobo.term.TermFrame):
            self.enrich_term(frame)
        elif isinstance(frame, fastobo.typedef.TypedefFrame):
            self.enrich_relationship(frame)
        else:
            warnings.warn(
                "cannot handle OBO instances", NotImplementedWarning, stacklevel=3
            )

    # --- Frame processing ---------------------------------------------------

    def extract_metadata(self, header: fastobo.header.HeaderFrame) -> Metadata:
        """Extract a `Metadata` from a `HeaderFrame`."""
        metadata = Metadata()
        for clause in header:
            process_clause_header(clause, metadata, self.ont)
        return metadata

    __non_one_clause = {k: attrgetter(k) for k in ("union_of", "intersection_of")}

    def enrich_term(self, frame: fastobo.term.TermFrame) -> Term:
        """Given a `TermFrame`, create a new `Term` or enrich an existing one."""
        # Create a term, or get the existing one if any
        id_ = str(frame.id)
        try:
            term = self.ont.get_term(id_)
        except KeyError:
            term = self.ont.create_term(id_)
        data = term._data()
        # Process all clauses in the frame
        for clause in frame:
            process_clause_term(clause, data, self.ont)
        # check cardinality of constrained clauses
        for attr, getter in self.__non_one_clause.items():
            if len(getter(data)) == 1:
                raise ValueError(f"{attr!r} cannot have a cardinality of 1")
        # return the enriched term
        return term

    def enrich_relationship(self, frame: fastobo.typedef.TypedefFrame) -> Relationship:
        """Given a `TypedefFrame`, create or enrich a `Relationship`."""
        # Create a relationship, or get the existing one if any
        id_ = str(frame.id)
        try:
            rship = self.ont.get_relationship(id_)
        except KeyError:
            rship = self.ont.create_relationship(id_)
        data = rship._data()
        # Process all clauses in the frame
        for clause in frame:
            process_clause_typedef(clause, data, self.ont)
        # check cardinality of constrained clauses
        for attr, getter in self.__non_one_clause.items():
            if len(getter(data)) == 1:
                raise ValueError(f"{attr!r} cannot have a cardinality of 1")
        # return the enriched relationship
        return rship


# --- Miscellaneous AST nodes ------------------------------------------------


def _extract_definition(clause: DefClause) -> Definition:
    return Definition(clause.definition, map(_extract_xref, clause.xrefs))


def _extract_property_value(pv: fastobo.pv.AbstractPropertyValue) -> PropertyValue:
    if isinstance(pv, fastobo.pv.LiteralPropertyValue):
        lpv = LiteralPropertyValue.__new__(LiteralPropertyValue)
        lpv.property = str(pv.relation)
        lpv.literal = pv.value
        lpv.datatype = str(pv.datatype)
        return lpv
    elif isinstance(pv, fastobo.pv.ResourcePropertyValue):
        rpv = ResourcePropertyValue.__new__(ResourcePropertyValue)
        rpv.property = str(pv.relation)
        rpv.resource = str(pv.value)
        return rpv
    else:
        msg = "'pv' must be AbstractPropertyValue, not {}"
        raise TypeError(msg.format(type(pv).__name__))


def _extract_synonym_data(syn: fastobo.syn.Synonym) -> SynonymData:
    xrefs = map(_extract_xref, syn.xrefs)
    type_ = str(syn.type) if syn.type is not None else None
    return SynonymData(syn.desc, syn.scope, type_, xrefs)


def _extract_xref(xref: fastobo.xref.Xref) -> Xref:
    x = Xref.__new__(Xref)
    x.id = str(xref.id)
    x.description = xref.desc
    return x


# --- Header clauses ---------------------------------------------------------


@functools.singledispatch
def process_clause_header(
    clause: fastobo.header.BaseHeaderClause, meta: Metadata, ont: Ontology
):
    raise TypeError(f"unexpected type: {type(clause).__name__}")


@process_clause_header.register(fastobo.header.AutoGeneratedByClause)
def _process_clause_header_auto_generated_by(clause, meta, ont):
    meta.auto_generated_by = clause.name


@process_clause_header.register(fastobo.header.DataVersionClause)
def _process_clause_header_data_version(clause, meta, ont):
    meta.data_version = clause.version


@process_clause_header.register(fastobo.header.DateClause)
def _process_clause_header_date(clause, meta, ont):
    meta.date = clause.date


@process_clause_header.register(fastobo.header.DefaultNamespaceClause)
def _process_clause_header_default_namespace(clause, meta, ont):
    meta.default_namespace = str(clause.namespace)


@process_clause_header.register(fastobo.header.FormatVersionClause)
def _process_clause_header_format_version(clause, meta, ont):
    meta.format_version = clause.version


@process_clause_header.register(fastobo.header.IdspaceClause)
def _process_clause_header_idspace(clause, meta, ont):
    meta.idspaces[str(clause.prefix)] = str(clause.url), clause.description


@process_clause_header.register(fastobo.header.ImportClause)
def _process_clause_header_import(clause, meta, ont):
    meta.imports.add(clause.reference)


@process_clause_header.register(fastobo.header.OntologyClause)
def _process_clause_header_ontology(clause, meta, ont):
    meta.ontology = clause.ontology


@process_clause_header.register(fastobo.header.OwlAxiomsClause)
def _process_clause_header_owl_axioms(clause, meta, ont):
    meta.owl_axioms.append(clause.axioms)


@process_clause_header.register(fastobo.header.RemarkClause)
def _process_clause_header_remark(clause, meta, ont):
    meta.remarks.add(clause.remark)


@process_clause_header.register(fastobo.header.SavedByClause)
def _process_clause_header_saved_by(clause, meta, ont):
    meta.saved_by = clause.name


@process_clause_header.register(fastobo.header.SubsetdefClause)
def _process_clause_header_subsetdef(clause, meta, ont):
    meta.subsetdefs.add(Subset(str(clause.subset), clause.description))


@process_clause_header.register(fastobo.header.SynonymTypedefClause)
def _process_clause_header_synonymtypedef(clause, meta, ont):
    meta.synonymtypedefs.add(
        SynonymType(str(clause.typedef), clause.description, clause.scope)
    )


@process_clause_header.register(fastobo.header.TreatXrefsAsEquivalentClause)
@process_clause_header.register(fastobo.header.TreatXrefsAsGenusDifferentiaClause)
@process_clause_header.register(fastobo.header.TreatXrefsAsHasSubclassClause)
@process_clause_header.register(fastobo.header.TreatXrefsAsIsAClause)
@process_clause_header.register(fastobo.header.TreatXrefsAsRelationshipClause)
@process_clause_header.register(
    fastobo.header.TreatXrefsAsReverseGenusDifferentiaClause
)
def _process_clause_header_treat_xrefs(clause, meta, ont):
    warnings.warn(
        f"cannot process `{clause}` macro", NotImplementedWarning, stacklevel=3
    )


@process_clause_header.register(fastobo.header.NamespaceIdRuleClause)
def _process_clause_header_namespace_id_rule(clause, meta, ont):
    meta.namespace_id_rule = clause.rule


@process_clause_header.register(fastobo.header.UnreservedClause)
def _process_clause_header_unreserved(clause, meta, ont):
    meta.unreserved.setdefault(clause.raw_tag(), set()).add(clause.raw_value())


# --- Term & Typedef clauses -------------------------------------------------


@functools.singledispatch
def process_clause_term(
    clause: fastobo.term.BaseTermClause, term: TermData, ont: Ontology
):
    """Populate the proper `term` field with data extracted from `clause`."""
    raise TypeError(f"unexpected type: {type(clause).__name__}")


@functools.singledispatch
def process_clause_typedef(
    clause: fastobo.typedef.BaseTypedefClause, rel: RelationshipData, ont: Ontology
):
    """Populate the proper `rel` field with data extracted from `clause`."""
    raise TypeError(f"unexpected type: {type(clause).__name__}")


@process_clause_term.register(fastobo.term.AltIdClause)
def _process_clause_term_alt_id(clause, entity, ont):
    id_ = str(clause.alt_id)
    entity.alternate_ids.add(id_)


@process_clause_typedef.register(fastobo.typedef.AltIdClause)
def _process_clause_typedef_alt_id(clause, entity, ont):
    id_ = str(clause.alt_id)
    entity.alternate_ids.add(id_)


@process_clause_term.register(fastobo.term.BuiltinClause)
@process_clause_typedef.register(fastobo.typedef.BuiltinClause)
def _process_clause_entity_builtin(clause, entity, ont):
    entity.builtin = clause.builtin


@process_clause_term.register(fastobo.term.CommentClause)
@process_clause_typedef.register(fastobo.typedef.CommentClause)
def _process_clause_entity_comment(clause, entity, ont):
    entity.comment = clause.comment


@process_clause_term.register(fastobo.term.ConsiderClause)
def _process_clause_term_consider(clause, entity, ont):
    entity.consider.add(str(clause.term))


@process_clause_typedef.register(fastobo.typedef.ConsiderClause)
def _process_clause_typedef_consider(clause, entity, ont):
    entity.consider.add(str(clause.typedef))


@process_clause_term.register(fastobo.term.CreatedByClause)
@process_clause_typedef.register(fastobo.typedef.CreatedByClause)
def _process_clause_entity_created_by(clause, entity, ont):
    entity.created_by = clause.creator


@process_clause_term.register(fastobo.term.CreationDateClause)
@process_clause_typedef.register(fastobo.typedef.CreationDateClause)
def _process_clause_entity_creation_date(clause, entity, ont):
    date = clause.date
    if isinstance(date, datetime.datetime):
        entity.creation_date = date
    else:
        warnings.warn(
            "source document contains incomplete creation date: {}".format(date.isoformat()),
            SyntaxWarning,
            stacklevel=3
        )
        entity.creation_date = datetime.datetime(date.year, date.month, date.day)


@process_clause_term.register(fastobo.term.DefClause)
@process_clause_typedef.register(fastobo.typedef.DefClause)
def _process_clause_entity_definition(clause, entity, ont):
    entity.definition = _extract_definition(clause)


@process_clause_term.register(fastobo.term.DisjointFromClause)
def _process_clause_term_disjoint_from(clause, entity, ont):
    entity.disjoint_from.add(str(clause.term))


@process_clause_typedef.register(fastobo.typedef.DisjointFromClause)
def _process_clause_typedef_disjoint_from(clause, entity, ont):
    entity.disjoint_from.add(str(clause.typedef))


@process_clause_typedef.register(fastobo.typedef.DisjointOverClause)
def _process_clause_typedef_disjoint_over(clause, entity, ont):
    entity.disjoint_over.add(str(clause.typedef))


@process_clause_typedef.register(fastobo.typedef.DomainClause)
def _process_clause_typedef_domain(clause, entity, ont):
    entity.domain = str(clause.domain)


@process_clause_term.register(fastobo.term.EquivalentToClause)
def _process_clause_term_equivalent_to(clause, entity, ont):
    entity.equivalent_to.add(str(clause.term))


@process_clause_typedef.register(fastobo.typedef.EquivalentToClause)
def _process_clause_typedef_equivalent_to(clause, entity, ont):
    entity.equivalent_to.add(str(clause.typedef))


@process_clause_typedef.register(fastobo.typedef.EquivalentToChainClause)
def _process_clause_typedef_equivalent_to_chain(clause, entity, ont):
    warnings.warn(
        f"cannot process `{clause}` macro", NotImplementedWarning, stacklevel=3
    )


@process_clause_typedef.register(fastobo.typedef.ExpandAssertionToClause)
def _process_clause_typedef_expand_assertion_to(clause, entity, ont):
    entity.expand_assertion_to.add(_extract_definition(clause))


@process_clause_typedef.register(fastobo.typedef.ExpandExpressionToClause)
def _process_clause_typedef_expand_expression_to(clause, entity, ont):
    entity.expand_expression_to.add(_extract_definition(clause))


@process_clause_typedef.register(fastobo.typedef.HoldsOverChainClause)
def _process_clause_typedef_holds_over_chain(clause, entity, ont):
    entity.holds_over_chain.add((str(clause.first), str(clause.last)))


@process_clause_term.register(fastobo.term.IntersectionOfClause)
def _process_clause_term_intersection_of(clause, entity, ont):
    if clause.typedef is None:
        entity.intersection_of.add(str(clause.term))
    else:
        entity.intersection_of.add((str(clause.typedef), str(clause.term)))


@process_clause_typedef.register(fastobo.typedef.IntersectionOfClause)
def _process_clause_typedef_intersection_of(clause, entity, ont):
    entity.intersection_of.add(str(clause.typedef))


@process_clause_typedef.register(fastobo.typedef.InverseOfClause)
def _process_clause_typedef_inverse_of(clause, entity, ont):
    entity.inverse_of = str(clause.typedef)


@process_clause_term.register(fastobo.term.IsAClause)
def _process_clause_term_is_a(clause, entity, ont):
    ont._terms.lineage[entity.id].sup.add(str(clause.term))


@process_clause_typedef.register(fastobo.typedef.IsAClause)
def _process_clause_typedef_is_a(clause, entity, ont):
    ont._relationships.lineage[entity.id].sup.add(str(clause.typedef))


@process_clause_term.register(fastobo.term.IsAnonymousClause)
@process_clause_typedef.register(fastobo.typedef.IsAnonymousClause)
def _process_clause_entity_is_anonymous(clause, entity, ont):
    entity.anonymous = clause.anonymous


@process_clause_typedef.register(fastobo.typedef.IsAntiSymmetricClause)
def _process_clause_typedef_is_anti_symmetric(clause, entity, ont):
    entity.antisymmetric = clause.anti_symmetric


@process_clause_typedef.register(fastobo.typedef.IsAsymmetricClause)
def _process_clause_typedef_is_asymmetric(clause, entity, ont):
    entity.asymmetric = clause.asymmetric


@process_clause_typedef.register(fastobo.typedef.IsClassLevelClause)
def _process_clause_typedef_is_class_level(clause, entity, ont):
    entity.class_level = clause.class_level


@process_clause_typedef.register(fastobo.typedef.IsCyclicClause)
def _process_clause_typedef_is_cyclic(clause, entity, ont):
    entity.cyclic = clause.cyclic


@process_clause_typedef.register(fastobo.typedef.IsFunctionalClause)
def _process_clause_typedef_is_functional(clause, entity, ont):
    entity.functional = clause.functional


@process_clause_typedef.register(fastobo.typedef.IsInverseFunctionalClause)
def _process_clause_typedef_is_inverse_functional(clause, entity, ont):
    entity.inverse_functional = clause.inverse_functional


@process_clause_typedef.register(fastobo.typedef.IsMetadataTagClause)
def _process_clause_typedef_is_metadata_tag(clause, entity, ont):
    entity.metadata_tag = clause.metadata_tag


@process_clause_typedef.register(fastobo.typedef.IsObsoleteClause)
def _process_clause_typedef_is_obsolete(clause, entity, ont):
    entity.obsolete = clause.obsolete


@process_clause_typedef.register(fastobo.typedef.IsReflexiveClause)
def _process_clause_typedef_is_reflexive(clause, entity, ont):
    entity.reflexive = clause.reflexive


@process_clause_typedef.register(fastobo.typedef.IsSymmetricClause)
def _process_clause_typedef_is_symmetric(clause, entity, ont):
    entity.symmetric = clause.symmetric


@process_clause_typedef.register(fastobo.typedef.IsTransitiveClause)
def _process_clause_typedef_is_transitive(clause, entity, ont):
    entity.transitive = clause.transitive


@process_clause_term.register(fastobo.term.IsObsoleteClause)
@process_clause_typedef.register(fastobo.typedef.IsObsoleteClause)
def _process_clause_entity_is_obsolete(clause, entity, ont):
    entity.obsolete = clause.obsolete


@process_clause_term.register(fastobo.term.NameClause)
@process_clause_typedef.register(fastobo.typedef.NameClause)
def _process_clause_entity_name(clause, entity, ont):
    entity.name = clause.name


@process_clause_term.register(fastobo.term.NamespaceClause)
@process_clause_typedef.register(fastobo.typedef.NamespaceClause)
def _process_clause_entity_namespace(clause, entity, ont):
    entity.namespace = str(clause.namespace)


@process_clause_header.register(fastobo.header.PropertyValueClause)
@process_clause_term.register(fastobo.term.PropertyValueClause)
@process_clause_typedef.register(fastobo.typedef.PropertyValueClause)
def _process_clause_entity_property_value(clause, entity, ont):
    entity.annotations.add(_extract_property_value(clause.property_value))


@process_clause_typedef.register(fastobo.typedef.RangeClause)
def _process_clause_typedef_range(clause, entity, ont):
    entity.range = str(clause.range)


@process_clause_term.register(fastobo.term.RelationshipClause)
def _process_clause_term_relationship(clause, entity, ont):
    entity.relationships.setdefault(str(clause.typedef), set()).add(str(clause.term))


@process_clause_typedef.register(fastobo.typedef.RelationshipClause)
def _process_clause_typedef_relationship(clause, entity, ont):
    entity.relationships.setdefault(str(clause.typedef), set()).add(str(clause.target))


@process_clause_term.register(fastobo.term.ReplacedByClause)
def _process_clause_term_replaced_by(clause, entity, ont):
    entity.replaced_by.add(str(clause.term))


@process_clause_typedef.register(fastobo.typedef.ReplacedByClause)
def _process_clause_typedef_replaced_by(clause, entity, ont):
    entity.replaced_by.add(str(clause.typedef))


@process_clause_term.register(fastobo.term.SubsetClause)
@process_clause_typedef.register(fastobo.typedef.SubsetClause)
def _process_clause_entity_subset(clause, entity, ont):
    entity.subsets.add(str(clause.subset))


@process_clause_term.register(fastobo.term.SynonymClause)
@process_clause_typedef.register(fastobo.typedef.SynonymClause)
def _process_clause_entity_synonym(clause, entity, ont):
    entity.synonyms.add(_extract_synonym_data(clause.synonym))


@process_clause_typedef.register(fastobo.typedef.TransitiveOverClause)
def _process_clause_typedef_transitive_over(clause, entity, ont):
    entity.transitive_over.add(str(clause.typedef))


@process_clause_term.register(fastobo.term.UnionOfClause)
def _process_clause_term_union_of(clause, entity, ont):
    entity.union_of.add(str(clause.term))


@process_clause_typedef.register(fastobo.typedef.UnionOfClause)
def _process_clause_typedef_union_of(clause, entity, ont):
    entity.union_of.add(str(clause.term))


@process_clause_term.register(fastobo.term.XrefClause)
@process_clause_typedef.register(fastobo.typedef.XrefClause)
def _process_clause_entity_xref(clause, entity, ont):
    entity.xrefs.add(_extract_xref(clause.xref))



import contextlib
import io
import itertools
import typing
import warnings
from os import PathLike, fspath
from typing import (
    BinaryIO,
    Container,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    Optional,
    Sized,
    Union,
)

from . import relationship
from .entity import Entity, EntityData
from .logic.lineage import Lineage
from .metadata import Metadata
from .relationship import Relationship, RelationshipData
from .synonym import SynonymType
from .term import Term, TermData
from .utils.io import decompress, get_handle, get_location
from .utils.iter import SizedIterator
from .utils.meta import roundrepr, typechecked

__all__ = ["Ontology"]
_D = typing.TypeVar("_D", bound=EntityData)


class _DataGraph(typing.Generic[_D], typing.Mapping[str, _D]):
    """A private data storage for a type of entity.

    This class is equivalent to a graph storing nodes in the ``entities``
    attribute, and directed edges corresponding to the sub-entity
    relationship between entities in the ``lineage`` attribute.
    """

    entities: MutableMapping[str, _D]
    lineage: MutableMapping[str, Lineage]

    def __init__(self, entities=None, lineage=None):
        self.entities = entities or {}
        self.lineage = lineage or {}

    def __contains__(self, key: object) -> bool:
        return key in self.entities

    def __len__(self) -> int:
        return len(self.entities)

    def __iter__(self):
        return iter(self.entities)

    def __getitem__(self, key: str) -> _D:
        return self.entities[key]


class _OntologyTerms(Sized, Container, Iterable[Term]):
    """A convenience wrapper over the terms of an ontology."""

    def __init__(self, ontology: "Ontology") -> None:
        self.__ontology = ontology

    def __len__(self) -> int:
        return sum(len(ref.terms()) for ref in self.__ontology.imports.values()) + len(
            self.__ontology._terms.entities
        )

    def __iter__(self) -> Iterator[Term]:
        return itertools.chain(
            (
                Term(self.__ontology, t._data())
                for ref in self.__ontology.imports.values()
                for t in ref.terms()
            ),
            (
                Term(self.__ontology, tdata)
                for tdata in self.__ontology._terms.entities.values()
            ),
        )

    def __contains__(self, item: object) -> bool:
        if isinstance(item, str):
            return (
                any(item in ref for ref in self.__ontology.imports.values())
                or item in self.__ontology._terms
            )
        return False


class _OntologyRelationships(Sized, Container, Iterable[Relationship]):
    """A convenience wrapper over the relationships of an ontology."""

    def __init__(self, ontology: "Ontology"):
        self.__ontology = ontology

    def __len__(self) -> int:
        return sum(
            len(ref.relationships()) for ref in self.__ontology.imports.values()
        ) + len(self.__ontology._relationships.entities)

    def __iter__(self) -> Iterator[Relationship]:
        return itertools.chain(
            (
                Relationship(self.__ontology, r._data())
                for ref in self.__ontology.imports.values()
                for r in ref.relationships()
            ),
            (
                Relationship(self.__ontology, rdata)
                for rdata in self.__ontology._relationships.entities.values()
            ),
        )

    def __contains__(self, item: object) -> bool:
        if isinstance(item, str):
            return (
                any(item in ref for ref in self.__ontology.imports.values())
                or item in self.__ontology._relationships
            )
        return False


class Ontology(Mapping[str, Union[Term, Relationship]]):
    """An ontology storing terms and the relationships between them.

    Ontologies can be loaded with ``pronto`` if they are serialized in any of
    the following ontology languages and formats at the moment:

    - `Ontology Web Language 2 <https://www.w3.org/TR/owl2-overview/>`_
      in `RDF/XML format
      <https://www.w3.org/TR/2012/REC-owl2-mapping-to-rdf-20121211/>`_.
    - `Open Biomedical Ontologies 1.4
      <http://owlcollab.github.io/oboformat/doc/obo-syntax.html>`_.
    - `OBO graphs <https://github.com/geneontology/obographs>`_ in
      `JSON <http://json.org/>`_ format.

    Attributes:
        metadata (Metadata): A data structure storing the metadata about the
            current ontology, either extracted from the ``owl:Ontology`` XML
            element or from the header of the OBO file.
        timeout (int): The timeout in seconds to use when performing network
            I/O, for instance when connecting to the OBO library to download
            imports. This is kept for reference, as it is not used after the
            initialization of the ontology.
        imports (~typing.Dict[str, Ontology]): A dictionary mapping references
            found in the import section of the metadata to resolved `Ontology`
            instances.

    """

    # Public attributes
    import_depth: int
    timeout: int
    imports: Dict[str, "Ontology"]
    path: Optional[str]
    handle: Optional[BinaryIO]

    # Private attributes
    _terms: _DataGraph[TermData]
    _relationships: _DataGraph[RelationshipData]

    # --- Constructors -------------------------------------------------------

    @classmethod
    def from_obo_library(
        cls,
        slug: str,
        import_depth: int = -1,
        timeout: int = 5,
        threads: Optional[int] = None,
    ) -> "Ontology":
        """Create an `Ontology` from a file in the OBO Library.

        This is basically just a shortcut constructor to avoid typing the full
        OBO Library URL each time.

        Arguments:
            slug (str): The filename of the ontology release to download from
                the OBO Library, including the file extension (should be one
                of ``.obo``, ``.owl`` or ``.json``).
            import_depth (int): The maximum depth of imports to resolve in the
                ontology tree. *Note that the library may not behave correctly
                when not importing the complete dependency tree, so you should
                probably use the default value and import everything*.
            timeout (int): The timeout in seconds to use when performing
                network I/O, for instance when connecting to the OBO library
                to download imports.
            threads (int): The number of threads to use when parsing, for
                parsers that support multithreading. Give `None` to autodetect
                the number of CPUs on the host machine.

        Example:
            >>> apo = pronto.Ontology.from_obo_library("apo.obo")
            >>> apo.metadata.ontology
            'apo'
            >>> apo.path
            'http://purl.obolibrary.org/obo/apo.obo'

        """
        return cls(
            f"http://purl.obolibrary.org/obo/{slug}", import_depth, timeout, threads
        )

    def __init__(
        self,
        handle: Union[BinaryIO, str, PathLike, None] = None,
        import_depth: int = -1,
        timeout: int = 5,
        threads: Optional[int] = None,
    ):
        """Create a new `Ontology` instance.

        Arguments:
            handle (str, ~typing.BinaryIO, ~os.PathLike, or None): Either the
                path to a file or a binary file handle that contains a
                serialized version of the ontology. If `None` is given, an
                empty `Ontology` is returned and can be populated manually.
            import_depth (int): The maximum depth of imports to resolve in the
                ontology tree. *Note that the library may not behave correctly
                when not importing the complete dependency tree, so you should
                probably use the default value and import everything*.
            timeout (int): The timeout in seconds to use when performing
                network I/O, for instance when connecting to the OBO library
                to download imports.
            threads (int): The number of threads to use when parsing, for
                parsers that support multithreading. Give `None` to autodetect
                the number of CPUs on the host machine.

        Raises:
            TypeError: When the given ``handle`` could not be used to parse
                and ontology.
            ValueError: When the given ``handle`` contains a serialized
                ontology not supported by any of the builtin parsers.

        """
        from .parsers import BaseParser

        with contextlib.ExitStack() as ctx:
            self.import_depth = import_depth
            self.timeout = timeout
            self.imports = dict()

            # self._inheritance = dict()
            # self._terms: Dict[str, TermData] = {}
            # self._relationships: Dict[str, RelationshipData] = {}
            self._terms = _DataGraph(entities={}, lineage={})
            self._relationships = _DataGraph(entities={}, lineage={})

            # Creating an ontology from scratch is supported
            if handle is None:
                self.metadata = Metadata()
                self.path = self.handle = None
                return

            # Get the path and the handle from arguments
            if isinstance(handle, (str, PathLike)):
                self.path = handle = fspath(handle)
                self.handle = ctx.enter_context(get_handle(handle, timeout))
                _handle = ctx.enter_context(decompress(self.handle))
                _detach = False
            elif hasattr(handle, "read"):
                self.path = get_location(handle)
                self.handle = handle
                _handle = decompress(self.handle)
                _detach = True
            else:
                raise TypeError(f"could not parse ontology from {handle!r}")

            # check value of `threads`
            if threads is not None and not threads > 0:
                raise ValueError("`threads` must be None or strictly positive")

            # Parse the ontology using the appropriate parser
            buffer = _handle.peek(io.DEFAULT_BUFFER_SIZE)
            for cls in BaseParser.__subclasses__():
                if cls.can_parse(typing.cast(str, self.path), buffer):
                    cls(self).parse_from(_handle)  # type: ignore
                    break
            else:
                raise ValueError(f"could not find a parser to parse {handle!r}")

            if _detach:
                _handle.detach()

    # --- Magic Methods ------------------------------------------------------

    def __len__(self) -> int:
        """Return the number of entities in the ontology.

        This method takes into accounts the terms and the relationships defined
        in the current ontology as well as all of its imports. To only count
        terms or relationships, use `len` on the iterator returned by the
        dedicated methods (e.g. ``len(ontology.terms())``).

        Example:
            >>> ms = pronto.Ontology.from_obo_library("ms.obo")
            >>> len(ms)
            6023
            >>> len(ms.terms())
            5995

        """
        return (
            len(self._terms.entities)
            + len(self._relationships.entities)
            + sum(map(len, self.imports.values()))
        )

    def __iter__(self) -> SizedIterator[str]:
        """Yield the identifiers of all the entities part of the ontology."""
        terms, relationships = self.terms(), self.relationships()
        entities: typing.Iterable[Entity] = itertools.chain(terms, relationships)
        return SizedIterator(
            (entity.id for entity in entities),
            length=len(terms) + len(relationships),
        )

    def __contains__(self, item: object) -> bool:
        if isinstance(item, str):
            # TODO: remove in v3.0.0
            if item in self._relationships or item in relationship._BUILTINS:
                warnings.warn(
                    "checking if an ontology contains a relationship with "
                    "`<id> in Ontology` will not be supported in future "
                    "versions, use `<id> in Ontology.relationships()`.",
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                return True
            return any(item in i for i in self.imports.values()) or item in self._terms
        return False

    def __getitem__(self, id: str) -> Union[Term, Relationship]:
        """Get any entity in the ontology graph with the given identifier."""
        # TODO: remove block in v3.0.0
        try:
            entity = self.get_relationship(id)
        except KeyError:
            pass
        else:
            warnings.warn(
                "indexing an ontology to retrieve a relationship will not be "
                "supported in future versions, use `Ontology.get_relationship` "
                "directly.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return entity

        try:
            return self.get_term(id)
        except KeyError:
            pass

        raise KeyError(id)

    def __repr__(self):
        """Return a textual representation of `self` that should roundtrip."""
        if self.path is not None:
            args = (self.path,)
        elif self.handle is not None:
            args = (self.handle,)
        else:
            args = ()
        kwargs = {"timeout": (self.timeout, 5)}
        if self.import_depth > 0:
            kwargs["import_depth"] = (self.import_depth, -1)
        return roundrepr.make("Ontology", *args, **kwargs)

    def __getstate__(self):
        state = self.__dict__.copy()
        state["handle"] = None
        return state

    def __setstate__(self, state):
        self.__dict__ = state

    # --- Serialization utils ------------------------------------------------

    def dump(self, file: BinaryIO, format: str = "obo"):
        """Serialize the ontology to a given file-handle.

        Arguments:
            file (~typing.BinaryIO): A binary file handle open in reading mode
                to write the serialized ontology into.
            format (str): The serialization format to use. Currently supported
                formats are: **obo**, **json**.

        Example:
            >>> ms = pronto.Ontology.from_obo_library("ms.obo")
            >>> with open("ms.json", "wb") as f:
            ...     ms.dump(f, format="json")

        """
        from .serializers import BaseSerializer

        for cls in BaseSerializer.__subclasses__():
            if cls.format == format:
                cls(self).dump(file)  # type: ignore
                break
        else:
            raise ValueError(f"could not find a serializer to handle {format!r}")

    def dumps(self, format: str = "obo") -> str:
        """Get a textual representation of the serialization ontology.

        Example:
            >>> go = pronto.Ontology("go.obo")
            >>> print(go.dumps())
            format-version: 1.2
            data-version: releases/2019-07-01
            ...

        """
        s = io.BytesIO()
        self.dump(s, format=format)
        return s.getvalue().decode("utf-8")

    # --- Data accessors -----------------------------------------------------

    def synonym_types(self) -> SizedIterator[SynonymType]:
        """Iterate over the synonym types of the ontology graph."""
        sources = [i.synonym_types() for i in self.imports.values()]
        sources.append(self.metadata.synonymtypedefs)  # type: ignore
        length = sum(map(len, sources))
        return SizedIterator(itertools.chain.from_iterable(sources), length)

    def terms(self) -> _OntologyTerms:
        """Query the terms of an ontology.

        Example:
            >>> pato = pronto.Ontology.from_obo_library("pato.obo")
            >>> len(pato.terms())
            2661
            >>> "PATO:0000186" in pato.terms()
            True
            >>> for term in sorted(pato.terms()):
            ...     print(term)
            Term('PATO:0000000', name='obsolete pato')
            Term('PATO:0000001', name='quality')
            ...

        """
        return _OntologyTerms(self)

    def relationships(self) -> _OntologyRelationships:
        """Query the relationships of an ontology.

        Example:
            >>> pato = pronto.Ontology.from_obo_library("pato.obo")
            >>> len(pato.relationships())
            24
            >>> "reciprocal_of" in pato.relationships()
            True
            >>> for relationship in pato.relationships():
            ...     print(relationship)
            Relationship('correlates_with', ...)
            Relationship('decreased_in_magnitude_relative_to', ...)
            ...

        """
        return _OntologyRelationships(self)

    @typechecked()
    def create_term(self, id: str) -> Term:
        """Create a new term with the given identifier.

        Returns:
            `Term`: the newly created term view, which attributes can the be
            modified directly.

        Raises:
            ValueError: if the provided ``id`` already identifies an entity
                in the ontology graph, or if it is not a valid OBO identifier.

        """
        if id in self:
            raise ValueError(f"identifier already in use: {id} ({self[id]})")
        self._terms.entities[id] = termdata = TermData(id)
        self._terms.lineage[id] = Lineage()
        return Term(self, termdata)

    @typechecked()
    def create_relationship(self, id: str) -> Relationship:
        """Create a new relationship with the given identifier.

        Raises:
            ValueError: if the provided ``id`` already identifies an entity
                in the ontology graph.

        """
        if id in self:
            raise ValueError(f"identifier already in use: {id} ({self[id]})")
        self._relationships.entities[id] = reldata = RelationshipData(id)
        self._relationships.lineage[id] = Lineage()
        return Relationship(self, reldata)

    @typechecked()
    def get_term(self, id: str) -> Term:
        """Get a term in the ontology graph from the given identifier.

        Raises:
            KeyError: if the provided ``id`` cannot be found in the terms of
                the ontology graph.

        """
        try:
            return Term(self, self._terms[id])
        except KeyError:
            pass
        for dep in self.imports.values():
            try:
                return Term(self, dep.get_term(id)._data())
            except KeyError:
                pass
        raise KeyError(id)

    @typechecked()
    def get_relationship(self, id: str) -> Relationship:
        """Get a relationship in the ontology graph from the given identifier.

        Builtin ontologies (``is_a`` and ``has_subclass``) can be accessed
        with this method.

        Raises:
            KeyError: if the provided ``id`` cannot be found in the
                relationships of the ontology graph.

        """
        # TODO: remove block in v3.0.0
        if id in relationship._BUILTINS:
            warnings.warn(
                "using the `is_a` relationship not be supported in future versions, "
                "use `superclasses` and `subclasses` API of entities instead.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return Relationship(self, relationship._BUILTINS[id])

        try:
            return Relationship(self, self._relationships[id])
        except KeyError:
            pass

        for dep in self.imports.values():
            try:
                return Relationship(self, dep.get_relationship(id)._data())
            except KeyError:
                pass

        raise KeyError(id)

    @typechecked()
    def get_synonym_type(self, id: str) -> SynonymType:
        """Get a synonym type in the ontology graph from the given identifier.

        Raises:
            KeyError: if the provided ``id`` does not resolve to a synonym type
                declared in this `Ontology` or one of its imports.

        """
        ty = next((ty for ty in self.metadata.synonymtypedefs if ty.id == id), None)
        if ty is not None:
            return ty
        for dep in self.imports.values():
            try:
                return dep.get_synonym_type(id)
            except KeyError:
                pass
        raise KeyError(id)



import collections
import contextlib
import functools
import inspect
import itertools
import sys
import typing
from typing import Callable, Iterator, List, Tuple, Type

T = typing.TypeVar("T")
F = typing.TypeVar("F", bound=Callable[..., object])


class typechecked(object):

    _disable = 0

    if sys.version_info >= (3, 7):
        Set = set
        FrozenSet = frozenset
    else:
        Set = typing.Set
        FrozenSet = typing.FrozenSet

    @classmethod
    @typing.no_type_check
    def check_type(cls, hint: object, value: object) -> Tuple[bool, str]:
        # None: check if None
        if hint is None.__class__:
            return (value is None, "None")
        # typing.Any is always true
        if typing.cast(str, getattr(hint, "_name", None)) == "typing.Any":
            return (True, "any object")
        # typing.Set needs to check member types
        if typing.cast(object, getattr(hint, "__origin__", None)) is cls.Set:
            if not isinstance(value, collections.abc.MutableSet):
                return (False, f"set of { hint.__args__ }")
            for arg in value:
                well_typed, type_name = cls.check_type(hint.__args__[0], arg)
                if not well_typed:
                    return (False, f"set of { type_name }")
            return (True, f"set of { hint.__args__ }")
        # typing.FrozenSet needs to check member types
        if typing.cast(object, getattr(hint, "__origin__", None)) is cls.FrozenSet:
            if not isinstance(value, collections.abc.Set):
                return (False, f"frozen set of { hint.__args__ }")
            for arg in value:
                well_typed, type_name = cls.check_type(hint.__args__[0], arg)
                if not well_typed:
                    return (False, f"frozenset of { type_name }")
            return (True, f"frozenset of { hint.__args__[0] }")
        # typing.Union needs to be a valid type
        if getattr(hint, "__origin__", None) is typing.Union:
            results = {}
            for arg in hint.__args__:
                results[arg] = cls.check_type(arg, value)
            ok = any(ok for ok, name in results.values())
            return (ok, ", ".join(name for ok, name in results.values()))
        # direct type annotation: simply do an instance check
        if isinstance(hint, type):
            return (isinstance(value, hint), hint.__name__)
        return (False, "<unknown>")

    @classmethod
    @contextlib.contextmanager
    def disabled(cls) -> Iterator[None]:
        cls._disable += 1
        try:
            yield
        finally:
            cls._disable -= 1

    def __init__(self, property: bool = False) -> None:
        self.property = property

    def __call__(self, func: F) -> F:
        if not __debug__:
            return func

        hints = typing.get_type_hints(func)
        signature = inspect.signature(func)

        @functools.wraps(func)
        def newfunc(*args, **kwargs):
            if not self._disable:
                callargs = signature.bind(*args, **kwargs).arguments
                for name, value in callargs.items():
                    if name in hints:
                        well_typed, type_name = self.check_type(hints[name], value)
                        if not well_typed:
                            msg = f"'{{}}' must be {type_name}, not {type(value).__name__}"
                            if self.property:
                                raise TypeError(msg.format(func.__name__))
                            else:
                                raise TypeError(msg.format(name))
            return func(*args, **kwargs)

        return newfunc  # type: ignore


class roundrepr(object):
    """A class-decorator to build a minimal `__repr__` method that roundtrips."""

    @staticmethod
    def make(class_name: str, *args: object, **kwargs: Tuple[object, object]) -> str:
        """Generate a repr string.

        Positional arguments should be the positional arguments used to
        construct the class. Keyword arguments should consist of tuples of
        the attribute value and default. If the value is the default, then
        it won't be rendered in the output.

        Example:
            >>> from pronto.utils.meta import roundrepr
            >>> class MyClass(object):
            ...     def __init__(self, name=None):
            ...         self.name = name
            ...     def __repr__(self):
            ...         return roundrepr.make('MyClass', 'foo', name=(self.name, None))
            >>> MyClass('Will')
            MyClass('foo', name='Will')
            >>> MyClass(None)
            MyClass('foo')

        Note:
            This functions uses code developed by `Will McGugan <https://github.com/willmcgugan>`_
            for `PyFilesystem2 <https://github.com/PyFilesystem/pyfilesystem2/blob/master/fs/_repr.py>`_

        """
        arguments: List[str] = [repr(arg) for arg in args]
        arguments.extend(
            [
                "{}={!r}".format(name, value)
                for name, (value, default) in sorted(kwargs.items())
                if value != default and value
            ]
        )
        return "{}({})".format(class_name, ", ".join(arguments))

    def __new__(self, cls: Type[T]) -> T:  # type: ignore
        obj = super().__new__(self)
        obj.__init__()
        if isinstance(cls, type):
            return obj(cls)
        return obj

    def __call__(self, cls: T) -> T:

        # Extract signature of `__init__`
        sig = inspect.signature(cls.__init__)  # type: ignore
        if not all(p.kind is p.POSITIONAL_OR_KEYWORD for p in sig.parameters.values()):
            raise TypeError(
                "cannot use `roundrepr` on a class with variadic `__init__`"
            )

        # Derive the __repr__ implementation
        def __repr__(self_):
            """Return a `repr` string that roundtrips. Computed by @roundrepr."""
            args, kwargs = [], {}
            for name, param in itertools.islice(sig.parameters.items(), 1, None):
                if param.default is inspect.Parameter.empty:
                    args.append(getattr(self_, name))
                else:
                    kwargs[name] = (getattr(self_, name), param.default)
            return self.make(cls.__name__, *args, **kwargs)

        # Hotpatch the class and return it
        cls.__repr__ = __repr__  # type: ignore
        return cls



import abc
import io
from typing import BinaryIO, ClassVar

from ..ontology import Ontology


class BaseSerializer(abc.ABC):

    format: ClassVar[str] = NotImplemented

    def __init__(self, ont: Ontology):
        self.ont = ont

    @abc.abstractmethod
    def dump(self, file: BinaryIO) -> None:
        return NotImplemented  # type: ignore

    def dumps(self) -> str:
        s = io.BytesIO()
        self.dump(s)
        return s.getvalue().decode("utf-8")



import abc
import collections
import typing
import warnings
from typing import AbstractSet, Deque, Optional, Set, Tuple, cast

from ..utils.meta import roundrepr, typechecked

if typing.TYPE_CHECKING:
    from ..entity import Entity
    from ..term import Term, TermSet, TermData
    from ..relationship import Relationship, RelationshipData, RelationshipSet
    from ..ontology import Ontology, _DataGraph

_E = typing.TypeVar("_E", bound="Entity")

# --- Storage ----------------------------------------------------------------


@roundrepr
class Lineage(object):
    """An internal type to store the superclasses and subclasses of a term.

    Used in `Ontology` to cache subclassing relationships between terms since
    only the superclassing relationships are explicitly declared in source
    documents.
    """

    __slots__ = ("sub", "sup")

    def __init__(
        self,
        sub: Optional[AbstractSet[str]] = None,
        sup: Optional[AbstractSet[str]] = None,
    ):
        self.sub: Set[str] = set(sub if sub is not None else ())  # type: ignore
        self.sup: Set[str] = set(sup if sup is not None else ())  # type: ignore

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Lineage):
            return self.sub == other.sub and self.sup == other.sup
        return False

    # `Lineage` is mutable so this is the explicit way to tell it's unhashable
    # (see https://docs.python.org/3/reference/datamodel.html#object.__hash__)
    __hash__ = None  # type: ignore

# create an empty Lineage object to serve as a default
Lineage._EMPTY: Lineage = Lineage()
Lineage._EMPTY.sub = Lineage._EMPTY.sup = frozenset()

# --- Abstract handlers ------------------------------------------------------


class LineageHandler(typing.Generic[_E], typing.Iterable[_E]):
    def __init__(
        self, entity: _E, distance: Optional[int] = None, with_self: bool = False
    ):
        self.entity = entity
        self.distance = distance
        self.with_self = with_self
        # TODO: API compatibilty with previous iterator (remove for v3.0.0)
        self._it: typing.Optional[typing.Iterator[_E]] = None

    def __next__(self) -> _E:
        if self._it is None:
            ty = type(self.entity).__name__
            warnings.warn(
                f"`{ty}.subclasses()` and `{ty}.superclasses()` will not "
                "return iterators in next major version, but iterables. "
                "Update your code to use `iter(...)` if needed.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            self._it = iter(self)
        return next(cast(typing.Iterator[_E], self._it))

    def _add(self, subclass: _E, superclass: _E):
        if superclass._ontology() is not subclass._ontology():
            ty = type(subclass).__name__
            raise ValueError(f"cannot use `{ty}` instances from different ontologies")
        lineage = self._get_data().lineage
        lineage[subclass.id].sup.add(superclass.id)
        lineage[superclass.id].sub.add(subclass.id)

    def _remove(self, subclass: _E, superclass: _E):
        if superclass._ontology() is not subclass._ontology():
            ty = type(subclass).__name__
            raise ValueError(f"cannot use `{ty}` instances from different ontologies")
        lineage = self._get_data().lineage
        lineage[subclass.id].sup.remove(superclass.id)
        lineage[superclass.id].sub.remove(subclass.id)

    @abc.abstractmethod
    def __iter__(self) -> "LineageIterator[_E]":
        return NotImplemented

    @abc.abstractmethod
    def _get_data(self) -> "_DataGraph":
        return NotImplemented  # type: ignore

    @abc.abstractmethod
    def add(self, other: _E) -> None:
        return NotImplemented  # type: ignore

    @abc.abstractmethod
    def remove(self, other: _E) -> None:
        return NotImplemented  # type: ignore

    @abc.abstractmethod
    def clear(self) -> None:
        return NotImplemented  # type: ignore


class TermHandler(LineageHandler["Term"]):
    @abc.abstractmethod
    def __iter__(self) -> "TermIterator":
        return NotImplemented

    def _get_data(self) -> "_DataGraph[TermData]":
        return self.entity._ontology()._terms

    def to_set(self) -> "TermSet":
        return self.__iter__().to_set()


class RelationshipHandler(LineageHandler["Relationship"]):
    @abc.abstractmethod
    def __iter__(self) -> "RelationshipIterator":
        return NotImplemented

    def _get_data(self) -> "_DataGraph[RelationshipData]":
        return self.entity._ontology()._relationships

    def to_set(self) -> "RelationshipSet":
        return self.__iter__().to_set()


class SuperentitiesHandler(LineageHandler):
    @abc.abstractmethod
    def __iter__(self) -> "SuperentitiesIterator":
        return NotImplemented

    def add(self, superclass: _E):
        self._add(subclass=self.entity, superclass=superclass)

    def remove(self, superclass: _E):
        self._remove(subclass=self.entity, superclass=superclass)

    def clear(self):
        lineage = self._get_data().lineage
        for subclass in lineage[self.entity.id].sup:
            lineage[subclass].sub.remove(self.entity.id)
        lineage[self.entity.id].sup.clear()


class SubentitiesHandler(LineageHandler):
    @abc.abstractmethod
    def __iter__(self) -> "SubentitiesIterator":
        return NotImplemented

    def add(self, subclass: _E):
        self._add(superclass=self.entity, subclass=subclass)

    def remove(self, subclass: _E):
        self._remove(superclass=self.entity, subclass=subclass)

    def clear(self):
        lineage = self._get_data().lineage
        for superclass in lineage[self.entity.id].sub:
            lineage[superclass].sup.remove(self.entity.id)
        lineage[self.entity.id].sub.clear()


# --- Concrete handlers ------------------------------------------------------


class SubclassesHandler(SubentitiesHandler, TermHandler):
    def __iter__(self) -> "SubclassesIterator":
        return SubclassesIterator(
            self.entity, distance=self.distance, with_self=self.with_self
        )


class SubpropertiesHandler(SubentitiesHandler, RelationshipHandler):
    def __iter__(self) -> "SubpropertiesIterator":
        return SubpropertiesIterator(
            self.entity, distance=self.distance, with_self=self.with_self
        )


class SuperclassesHandler(SuperentitiesHandler, TermHandler):
    def __iter__(self) -> "SuperclassesIterator":
        return SuperclassesIterator(
            self.entity, distance=self.distance, with_self=self.with_self
        )


class SuperpropertiesHandler(SuperentitiesHandler, RelationshipHandler):
    def __iter__(self) -> "SuperpropertiesIterator":
        return SuperpropertiesIterator(
            self.entity, distance=self.distance, with_self=self.with_self
        )


# --- Abstract iterators -----------------------------------------------------


class LineageIterator(typing.Generic[_E], typing.Iterator[_E]):

    _distmax: float
    _ontology: Optional["Ontology"]
    _linked: Set[str]
    _done: Set[str]
    _frontier: Deque[Tuple[str, int]]
    _queue: Deque[str]

    # ---

    @abc.abstractmethod
    def _get_data(self) -> "_DataGraph":
        return NotImplemented  # type: ignore

    @abc.abstractmethod
    def _get_neighbors(self, node: str) -> Set[str]:
        return NotImplemented  # type: ignore

    @abc.abstractmethod
    def _get_entity(self, node: str) -> _E:
        return NotImplemented  # type: ignore

    @abc.abstractmethod
    def _maxlen(self) -> int:
        return NotImplemented  # type: ignore

    # ---

    def __init__(
        self, *entities: _E, distance: Optional[int] = None, with_self: bool = True
    ) -> None:

        self._distmax = float("inf") if distance is None else distance

        # if not term is given, `__next__` will raise `StopIterator` on
        # the first call without ever accessing `self._ontology`, so it's
        # safe not to initialise it here in that case.
        if entities:
            self._ontology = entities[0]._ontology()
        else:
            self._ontology = None

        self._linked: Set[str] = set()
        self._done: Set[str] = set()
        self._frontier: Deque[Tuple[str, int]] = collections.deque()
        self._queue: Deque[str] = collections.deque()

        for entity in entities:
            self._frontier.append((entity.id, 0))
            self._linked.add(entity.id)
            if with_self:
                self._queue.append(entity.id)

    def __iter__(self) -> "LineageIterator[_E]":
        return self

    def __length_hint__(self) -> int:
        """Return an estimate of the number of remaining entities to yield."""
        if self._queue or self._frontier:
            return self._maxlen() - len(self._linked) + len(self._queue)
        else:
            return 0

    def _next_id(self) -> Optional[str]:
        while self._frontier or self._queue:
            # Return any element currently queued
            if self._queue:
                return self._queue.popleft()
            # Get the next node in the frontier
            node, distance = self._frontier.popleft()
            self._done.add(node)
            # Process its neighbors if they are not too far
            neighbors: Set[str] = self._get_neighbors(node)
            if neighbors and distance < self._distmax:
                for node in sorted(neighbors.difference(self._done)):
                    self._frontier.append((node, distance + 1))
                for neighbor in sorted(neighbors.difference(self._linked)):
                    self._linked.add(neighbor)
                    self._queue.append(neighbor)
        # Stop iteration if no more elements to process
        return None

    def __next__(self) -> "_E":
        id_ = self._next_id()
        if id_ is None:
            raise StopIteration
        return self._get_entity(id_)


class TermIterator(LineageIterator["Term"]):
    def _maxlen(self):
        return len(self._ontology.terms())

    def _get_entity(self, id):
        return self._ontology.get_term(id)

    def _get_data(self):
        return self._ontology._terms

    def to_set(self) -> "TermSet":
        """Collect all classes into a `~pronto.TermSet`.

        Hint:
            This method is useful to query an ontology using a method chaining
            syntax, for instance::

            >>> cio = pronto.Ontology("cio.obo")
            >>> sorted(cio['CIO:0000034'].subclasses().to_set().ids)
            ['CIO:0000034', 'CIO:0000035', 'CIO:0000036']

        """
        from ..term import TermSet

        with typechecked.disabled():
            term_set = TermSet()
            term_set._ontology = self._ontology
            term_set._ids.update(iter(self._next_id, None))
            return term_set


class RelationshipIterator(LineageIterator["Relationship"]):
    def _maxlen(self):
        return len(self._ontology.relationships())

    def _get_entity(self, id):
        return self._ontology.get_relationship(id)

    def _get_data(self):
        return self._ontology._relationships

    def to_set(self) -> "RelationshipSet":
        """Collect all relationshisp into a `~pronto.RelationshipSet`.

        Hint:
            This method is useful to query an ontology using a method chaining
            syntax.

        """
        from ..relationship import RelationshipSet

        with typechecked.disabled():
            relationship_set = RelationshipSet()
            relationship_set._ontology = self._ontology
            relationship_set._ids.update(iter(self._next_id, None))
            return relationship_set


class SubentitiesIterator(LineageIterator):
    def _get_neighbors(self, node: str) -> Set[str]:
        return self._get_data().lineage.get(node, Lineage._EMPTY).sub


class SuperentitiesIterator(LineageIterator):
    def _get_neighbors(self, node: str) -> Set[str]:
        return self._get_data().lineage.get(node, Lineage._EMPTY).sup


# --- Concrete iterators -----------------------------------------------------


class SubclassesIterator(SubentitiesIterator, TermIterator):
    """An iterator over the subclasses of one or several `~pronto.Term`."""


class SuperclassesIterator(SuperentitiesIterator, TermIterator):
    """An iterator over the superclasses of one or several `~pronto.Term`."""


class SubpropertiesIterator(SubentitiesIterator, RelationshipIterator):
    """An iterator over the subproperties of one or several `~pronto.Relationship`."""


class SuperpropertiesIterator(SuperentitiesIterator, RelationshipIterator):
    """An iterator over the superproperties of one or several `~pronto.Relationship`."""
